unit UExecutor;

{$mode delphi}

interface

uses
  Classes, SysUtils, USource, Contnrs;

type
  TExecutorFunctions = class;

  TExecutorVariable = class
    Variable: TVariable;
    Value: TValue;
  end;

  { TExecutorVariables }

  TExecutorVariables = class(TObjectList)
    function SearchByVariable(Variable: TVariable): TExecutorVariable;
    function AddNew(Variable: TVariable): TExecutorVariable;
  end;

  { TExecutorType }

  TExecutorType = class
    TypeRef: TType;
    Functions: TExecutorFunctions;
    constructor Create;
    destructor Destroy; override;
  end;

  { TExecutorTypes }

  TExecutorTypes = class(TObjectList)
    function SearchByType(TypeRef: TType): TExecutorType;
    function AddNew(TypeRef: TType): TExecutorType;
  end;

  TExecutorFunctionCallback = function(Params: array of TValue): TValue of object;

  TExecutorFunction = class
    FunctionDef: TFunction;
    Callback: TExecutorFunctionCallback;
  end;

  { TExecutorFunctions }

  TExecutorFunctions = class(TObjectList)
    function SearchByFunction(FunctionDef: TFunction): TExecutorFunction;
    function AddNew(FunctionDef: TFunction): TExecutorFunction;
  end;

    { TExecutorBlock }

  TExecutorBlock = class
    Parent: TExecutorBlock;
    Types: TExecutorTypes;
    Variables: TExecutorVariables;
    Functions: TExecutorFunctions;
    function GetFunction(FunctionDef: TFunction): TExecutorFunction;
    function GetType(TypeDef: TType): TExecutorType;
    function GetVariable(Variable: TVariable): TExecutorVariable;
    function GetTypeFunction(TypeDef: TType; FunctionDef: TFunction): TExecutorFunction; overload;
    function GetTypeFunction(TypeDef: TType; FunctionName: string): TExecutorFunction; overload;
    constructor Create;
    destructor Destroy; override;
  end;

  TOutputEvent = procedure (Text: string) of object;

  { TExecutor }

  TExecutor = class
  private
    FOnOutput: TOutputEvent;
    SystemBlock: TExecutorBlock;
    function ExecuteWriteLn(Params: array of TValue): TValue;
    function ExecuteWrite(Params: array of TValue): TValue;
    function ExecuteIntToStr(Params: array of TValue): TValue;
    function ExecuteStrToInt(Params: array of TValue): TValue;
    function ExecuteStringAssign(Params: array of TValue): TValue;
    function ExecuteStringAdd(Params: array of TValue): TValue;
    function ExecuteStringEqual(Params: array of TValue): TValue;
    function ExecuteStringNotEqual(Params: array of TValue): TValue;
    function ExecuteIntegerAssign(Params: array of TValue): TValue;
    function ExecuteIntegerAdd(Params: array of TValue): TValue;
    function ExecuteIntegerSub(Params: array of TValue): TValue;
    function ExecuteIntegerEqual(Params: array of TValue): TValue;
    function ExecuteIntegerNotEqual(Params: array of TValue): TValue;
    procedure InitExecutorBlock(ExecutorBlock: TExecutorBlock; Block: TBlock);
  public
    Prog: TProgram;
    procedure ExecuteProgram(Prog: TProgram);
    procedure ExecuteBeginEnd(Block: TExecutorBlock; BeginEnd: TBeginEnd);
    procedure ExecuteCommand(Block: TExecutorBlock; Command: TCommand);
    procedure ExecuteIfThenElse(Block: TExecutorBlock; IfThenElse: TIfThenElse);
    procedure ExecuteWhileDo(Block: TExecutorBlock; WhileDo: TWhileDo);
    procedure ExecuteForToDo(Block: TExecutorBlock; ForToDo: TForToDo);
    procedure ExecuteBlock(ParentBlock: TExecutorBlock;Block: TBlock);
    function ExecuteFunctionCall(Block: TExecutorBlock; FunctionCall: TFunctionCall): TValue;
    procedure ExecuteAssignment(Block: TExecutorBlock; Assignment: TAssignment);
    function ExecuteExpression(Block: TExecutorBlock; Expression: TExpression): TValue;
    function ExecuteExpressionOperation(Block: TExecutorBlock; Expression: TExpressionOperation): TValue;
    function ExecuteExpressionOperand(Block: TExecutorBlock; Expression: TExpressionOperand): TValue;
    procedure Run;
    procedure Output(Text: string);
    property OnOutput: TOutputEvent read FOnOutput write FOnOutput;
  end;


implementation

{ TExecutorType }

constructor TExecutorType.Create;
begin
  Functions := TExecutorFunctions.Create;
end;

destructor TExecutorType.Destroy;
begin
  Functions.Free;
  inherited Destroy;
end;

{ TExecutorTypes }

function TExecutorTypes.SearchByType(TypeRef: TType): TExecutorType;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TExecutorType(Items[I]).TypeRef <> TypeRef) do Inc(I);
  if I < Count then Result := TExecutorType(Items[I])
    else Result := nil;
end;

function TExecutorTypes.AddNew(TypeRef: TType): TExecutorType;
begin
  Result := TExecutorType.Create;
  Result.TypeRef := TypeRef;
  Add(Result);
end;

{ TExecutorFunctions }

function TExecutorFunctions.SearchByFunction(FunctionDef: TFunction
  ): TExecutorFunction;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TExecutorFunction(Items[I]).FunctionDef <> FunctionDef) do Inc(I);
  if I < Count then Result := TExecutorFunction(Items[I])
    else Result := nil;
end;

function TExecutorFunctions.AddNew(FunctionDef: TFunction): TExecutorFunction;
begin
  Result := TExecutorFunction.Create;
  Result.FunctionDef := FunctionDef;
  Add(Result);
end;

{ TExecutorVariables }

function TExecutorVariables.SearchByVariable(Variable: TVariable): TExecutorVariable;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TExecutorVariable(Items[I]).Variable <> Variable) do Inc(I);
  if I < Count then Result := TExecutorVariable(Items[I])
    else Result := nil;
end;

function TExecutorVariables.AddNew(Variable: TVariable): TExecutorVariable;
begin
  Result := TExecutorVariable.Create;
  Result.Variable := Variable;
  Add(Result);
end;

{ TExecutorBlock }

function TExecutorBlock.GetFunction(FunctionDef: TFunction): TExecutorFunction;
begin
  Result := Functions.SearchByFunction(FunctionDef);
  if not Assigned(Result) and Assigned(Parent) then
    Result := Parent.GetFunction(FunctionDef);
end;

function TExecutorBlock.GetType(TypeDef: TType): TExecutorType;
begin
  Result := Types.SearchByType(TypeDef);
  if not Assigned(Result) and Assigned(Parent) then
    Result := Parent.GetType(TypeDef);
end;

function TExecutorBlock.GetVariable(Variable: TVariable): TExecutorVariable;
begin
  Result := Variables.SearchByVariable(Variable);
  if not Assigned(Result) and Assigned(Parent) then
    Result := Parent.GetVariable(Variable);
end;

function TExecutorBlock.GetTypeFunction(TypeDef: TType; FunctionDef: TFunction
  ): TExecutorFunction;
var
  ExecutorType: TExecutorType;
begin
  ExecutorType := GetType(TypeDef);
  Result := ExecutorType.Functions.SearchByFunction(FunctionDef);
end;

function TExecutorBlock.GetTypeFunction(TypeDef: TType; FunctionName: string
  ): TExecutorFunction;
begin
  Result := GetTypeFunction(TypeDef, TypeDef.Functions.SearchByName(FunctionName));
end;

constructor TExecutorBlock.Create;
begin
  Types := TExecutorTypes.Create;
  Variables := TExecutorVariables.Create;
  Functions := TExecutorFunctions.Create;
end;

destructor TExecutorBlock.Destroy;
begin
  Variables.Free;
  Functions.Free;
  Types.Free;
  inherited Destroy;
end;

{ TExecutor }

function TExecutor.ExecuteWriteLn(Params: array of TValue): TValue;
var
  I: Integer;
  Text: string;
begin
  Result := nil;
  Text := '';
  for I := 0 to Length(Params) - 1 do
    Text := Text + TValueString(Params[I]).Value;
  Output(Text + LineEnding);
end;

function TExecutor.ExecuteWrite(Params: array of TValue): TValue;
var
  I: Integer;
  Text: string;
begin
  Result := nil;
  Text := '';
  for I := 0 to Length(Params) - 1 do
    Text := Text + TValueString(Params[I]).Value;
  Output(Text);
end;

function TExecutor.ExecuteIntToStr(Params: array of TValue): TValue;
begin
  Result := TValueString.Create;
  TValueString(Result).Value := IntToStr(TValueInteger(Params[0]).Value);
end;

function TExecutor.ExecuteStrToInt(Params: array of TValue): TValue;
begin
  Result := TValueInteger.Create;
  TValueInteger(Result).Value := StrToInt(TValueString(Params[0]).Value);
end;

function TExecutor.ExecuteStringAssign(Params: array of TValue): TValue;
begin
  Result := TValueString.Create;
  TValueString(Result).Value := TValueString(Params[0]).Value;
end;

function TExecutor.ExecuteStringAdd(Params: array of TValue): TValue;
begin
  Result := TValueString.Create;
  TValueString(Result).Value := TValueString(Params[0]).Value + TValueString(Params[1]).Value;
end;

function TExecutor.ExecuteStringEqual(Params: array of TValue): TValue;
begin
  Result := TValueBoolean.Create;
  TValueBoolean(Result).Value := TValueString(Params[0]).Value = TValueString(Params[1]).Value;
end;

function TExecutor.ExecuteStringNotEqual(Params: array of TValue): TValue;
begin
  Result := TValueBoolean.Create;
  TValueBoolean(Result).Value := TValueString(Params[0]).Value <> TValueString(Params[1]).Value;
end;

function TExecutor.ExecuteIntegerAssign(Params: array of TValue): TValue;
begin
  Result := TValueInteger.Create;
  TValueInteger(Result).Value := TValueInteger(Params[0]).Value;
end;

function TExecutor.ExecuteIntegerAdd(Params: array of TValue): TValue;
begin
  Result := TValueInteger.Create;
  TValueInteger(Result).Value := TValueInteger(Params[0]).Value + TValueInteger(Params[1]).Value;
end;

function TExecutor.ExecuteIntegerSub(Params: array of TValue): TValue;
begin
  Result := TValueInteger.Create;
  TValueInteger(Result).Value := TValueInteger(Params[0]).Value - TValueInteger(Params[1]).Value;
end;

function TExecutor.ExecuteIntegerEqual(Params: array of TValue): TValue;
begin
  Result := TValueBoolean.Create;
  TValueBoolean(Result).Value := TValueInteger(Params[0]).Value = TValueInteger(Params[1]).Value;
end;

function TExecutor.ExecuteIntegerNotEqual(Params: array of TValue): TValue;
begin
  Result := TValueBoolean.Create;
  TValueBoolean(Result).Value := TValueInteger(Params[0]).Value <> TValueInteger(Params[1]).Value;
end;

procedure TExecutor.InitExecutorBlock(ExecutorBlock: TExecutorBlock; Block: TBlock);
var
  I: Integer;
  J: Integer;
  ExecutorFunction: TExecutorFunction;
  ExecutorType: TExecutorType;
begin
  for I := 0 to Block.Types.Count - 1 do begin
    ExecutorType := ExecutorBlock.Types.AddNew(TType(Block.Types[I]));
    for J := 0 to ExecutorType.TypeRef.Functions.Count - 1 do begin
      ExecutorFunction := ExecutorType.Functions.AddNew(TFunction(ExecutorType.TypeRef.Functions[J]));
      if ExecutorType.TypeRef.Name = 'string' then begin
        if ExecutorFunction.FunctionDef.Name = '_Assign' then begin
          ExecutorFunction.Callback := ExecuteStringAssign;
        end else
        if ExecutorFunction.FunctionDef.Name = '_Add' then begin
          ExecutorFunction.Callback := ExecuteStringAdd;
        end else
        if ExecutorFunction.FunctionDef.Name = '_Equal' then begin
          ExecutorFunction.Callback := ExecuteStringEqual;
        end;
        if ExecutorFunction.FunctionDef.Name = '_NotEqual' then begin
          ExecutorFunction.Callback := ExecuteStringNotEqual;
        end;
      end;
      if ExecutorType.TypeRef.Name = 'Integer' then begin
        if ExecutorFunction.FunctionDef.Name = '_Assign' then begin
          ExecutorFunction.Callback := ExecuteIntegerAssign;
        end else
        if ExecutorFunction.FunctionDef.Name = '_Add' then begin
          ExecutorFunction.Callback := ExecuteIntegerAdd;
        end else
        if ExecutorFunction.FunctionDef.Name = '_Sub' then begin
          ExecutorFunction.Callback := ExecuteIntegerSub;
        end else
        if ExecutorFunction.FunctionDef.Name = '_Equal' then begin
          ExecutorFunction.Callback := ExecuteIntegerEqual;
        end else
        if ExecutorFunction.FunctionDef.Name = '_NotEqual' then begin
          ExecutorFunction.Callback := ExecuteIntegerNotEqual;
        end;
      end;
    end;
  end;
  for I := 0 to Block.Variables.Count - 1 do
    ExecutorBlock.Variables.AddNew(TVariable(Block.Variables[I]));
  for I := 0 to Block.Functions.Count - 1 do begin
    ExecutorFunction := ExecutorBlock.Functions.AddNew(TFunction(Block.Functions[I]));
    if ExecutorFunction.FunctionDef.Name = 'Write' then begin
      ExecutorFunction.Callback := ExecuteWrite;
    end else
    if ExecutorFunction.FunctionDef.Name = 'WriteLn' then begin
      ExecutorFunction.Callback := ExecuteWriteLn;
    end;
    if ExecutorFunction.FunctionDef.Name = 'IntToStr' then begin
      ExecutorFunction.Callback := ExecuteIntToStr;
    end else
    if ExecutorFunction.FunctionDef.Name = 'StrToInt' then begin
      ExecutorFunction.Callback := ExecuteStrToInt;
    end;
  end;
end;

procedure TExecutor.ExecuteProgram(Prog: TProgram);
begin
  SystemBlock := TExecutorBlock.Create;
  InitExecutorBlock(SystemBlock, Prog.SystemBlock);
  ExecuteBlock(SystemBlock, Prog.Block);
  SystemBlock.Free;
end;

procedure TExecutor.ExecuteBeginEnd(Block: TExecutorBlock; BeginEnd: TBeginEnd);
var
  I: Integer;
begin
  for I := 0 to BeginEnd.Commands.Count - 1 do
    ExecuteCommand(Block, TCommand(BeginEnd.Commands[I]));
end;

procedure TExecutor.ExecuteCommand(Block: TExecutorBlock; Command: TCommand);
begin
  if Command is TBeginEnd then ExecuteBeginEnd(Block, TBeginEnd(Command))
  else if Command is TFunctionCall then ExecuteFunctionCall(Block, TFunctionCall(Command))
  else if Command is TAssignment then ExecuteAssignment(Block, TAssignment(Command))
  else if Command is TIfThenElse then ExecuteIfThenElse(Block, TIfThenElse(Command))
  else if Command is TWhileDo then ExecuteWhileDo(Block, TWhileDo(Command))
  else if Command is TForToDo then ExecuteForToDo(Block, TForToDo(Command))
  else raise Exception.Create('Unsupported command type');
end;

procedure TExecutor.ExecuteIfThenElse(Block: TExecutorBlock;
  IfThenElse: TIfThenElse);
var
  Value: TValue;
begin
  Value := ExecuteExpression(Block, IfThenElse.Expression);
  if Value is TValueBoolean then begin
    if TValueBoolean(Value).Value then ExecuteCommand(Block, IfThenElse.CommandThen)
      else ExecuteCommand(Block, IfThenElse.CommandElse);
  end else raise Exception.Create('Expected boolean value.');
end;

procedure TExecutor.ExecuteWhileDo(Block: TExecutorBlock; WhileDo: TWhileDo);
var
  Value: TValue;
begin
  while True do begin
    Value := ExecuteExpression(Block, WhileDo.Expression);
    if Value is TValueBoolean then begin
      if not TValueBoolean(Value).Value then Break;
      ExecuteCommand(Block, WhileDo.Command);
    end else raise Exception.Create('Expected boolean value.');
  end;
end;

procedure TExecutor.ExecuteForToDo(Block: TExecutorBlock; ForToDo: TForToDo);
var
  Value: TValue;
  Variable: TExecutorVariable;
  Limit: TValue;
begin
  Variable := Block.GetVariable(ForToDo.VariableRef);
  Variable.Value := ExecuteExpression(Block, ForToDo.ExpressionFrom);
  Limit := ExecuteExpression(Block, ForToDo.ExpressionTo);
  while True do begin
    ExecuteCommand(Block, ForToDo.Command);
    TValueInteger(Variable.Value).Value := TValueInteger(Variable.Value).Value + 1;
    if TValueInteger(Variable.Value).Value > TValueInteger(Limit).Value then Break;
  end;
end;

procedure TExecutor.ExecuteBlock(ParentBlock: TExecutorBlock; Block: TBlock);
var
  ExecutorBlock: TExecutorBlock;
begin
  ExecutorBlock := TExecutorBlock.Create;
  ExecutorBlock.Parent := ParentBlock;
  InitExecutorBlock(ExecutorBlock, Block);
  ExecuteBeginEnd(ExecutorBlock, Block.BeginEnd);
  ExecutorBlock.Free;
end;

function TExecutor.ExecuteFunctionCall(Block: TExecutorBlock;
  FunctionCall: TFunctionCall): TValue;
var
  ExecutorFunction: TExecutorFunction;
  Params: array of TValue;
  I: Integer;
begin
  Result := nil;
  ExecutorFunction := Block.GetFunction(FunctionCall.FunctionDef);
  if Assigned(ExecutorFunction) then begin
    SetLength(Params, FunctionCall.Params.Count);
    for I := 0 to FunctionCall.Params.Count - 1 do begin
      Params[I] := ExecuteExpression(Block, TExpression(FunctionCall.Params[0]));
    end;
    Result := ExecutorFunction.Callback(Params);
  end else raise Exception.Create('No executor for ' + FunctionCall.FunctionDef.Name + ' function.');
end;

procedure TExecutor.ExecuteAssignment(Block: TExecutorBlock;
  Assignment: TAssignment);
var
  Value: TValue;
  Variable: TExecutorVariable;
  ExecutorFunction: TExecutorFunction;
  Params: array of TValue;
begin
  Value := ExecuteExpression(Block, Assignment.Expression);
  Variable := Block.GetVariable(Assignment.Variable);
  ExecutorFunction := Block.GetTypeFunction(Assignment.Variable.TypeRef, '_Assign');
  if Assignment.Variable.TypeRef = Assignment.Expression.GetType then begin;
    SetLength(Params, 1);
    Params[0] := Value;
    Variable.Value := ExecutorFunction.Callback(Params);
  end else raise Exception('Assignment result type is ' + Variable.Variable.TypeRef.Name +
    ' but value is ' + Assignment.Expression.GetType.Name + '.');
end;

function TExecutor.ExecuteExpression(Block: TExecutorBlock;
  Expression: TExpression): TValue;
begin
  if Expression is TExpressionOperation then
    Result := ExecuteExpressionOperation(Block, TExpressionOperation(Expression))
  else
  if Expression is TExpressionOperand then
    Result := ExecuteExpressionOperand(Block, TExpressionOperand(Expression))
  else raise Exception.Create('Unknown expression class.');
end;

function TExecutor.ExecuteExpressionOperation(Block: TExecutorBlock;
  Expression: TExpressionOperation): TValue;
var
  I: Integer;
  Value: TValue;
  ExecutorFunction: TExecutorFunction;
  Params: array of TValue;
  FuncName: string;
begin
  if Expression.Operation = eoAdd then FuncName := '_Add'
  else if Expression.Operation = eoSub then FuncName := '_Sub'
  else if Expression.Operation = eoEqual then FuncName := '_Equal'
  else if Expression.Operation = eoNotEqual then FuncName := '_NotEqual'
  else raise Exception.Create('Unsupported operation type.');

  ExecutorFunction := Block.GetTypeFunction(Expression.TypeRef, FuncName);
  Result := Expression.TypeRef.ValueClass.Create;

  SetLength(Params, Expression.Items.Count);
  for I := 0 to Expression.Items.Count - 1 do begin
    Value := ExecuteExpression(Block, TExpression(Expression.Items[I]));
    Params[I] := Value;
  end;
  Result := ExecutorFunction.Callback(Params);
end;

function TExecutor.ExecuteExpressionOperand(Block: TExecutorBlock;
  Expression: TExpressionOperand): TValue;
begin
  if Assigned(Expression.VariableRef) then begin
    Result := Block.Variables.SearchByVariable(Expression.VariableRef).Value;
  end else
  if Assigned(Expression.ConstantRef) then begin
    Result := Expression.ConstantRef.Value;
  end else
  if Assigned(Expression.FunctionCall) then begin
    Result := ExecuteFunctionCall(Block, Expression.FunctionCall);
  end else raise Exception.Create('Unsupported exception operand type.');
end;

procedure TExecutor.Run;
begin
  ExecuteProgram(Prog);
end;

procedure TExecutor.Output(Text: string);
begin
  if Assigned(FOnOutput) then
    FOnOutput(Text);
end;


end.

