unit UParser;

{$mode delphi}{$H+}

interface

uses
  Classes, SysUtils, Contnrs, UTokenizer, USource;

type

  { TParser }

  TParser = class
  private
    FOnError: TErrorEvent;
    Tokenizer: TTokenizer;
    function ParseBeginEnd(Block: TBlock; out BeginEnd: TBeginEnd): Boolean;
    function ParseFunctionCall(Block: TBlock; out FunctionCall: TFunctionCall): Boolean;
    function ParseCommand(Block: TBlock; out Command: TCommand): Boolean;
    function ParseProgram(SystemBlock: TBlock; out Prog: TProgram): Boolean;
    function ParseBlock(ParentBlock: TBlock; out Block: TBlock): Boolean;
    function ParseBlockVar(Block: TBlock): Boolean;
    function ParseBlockConst(Block: TBlock): Boolean;
    function ParseAssignment(Block: TBlock; out Assignment: TAssignment): Boolean;
    function ParseExpression(Block: TBlock; out Expression: TExpression): Boolean;
    function ParseExpressionOperation(Block: TBlock; out ExpressionOperation: TExpressionOperation): Boolean;
    function ParseExpressionOperand(Block: TBlock; out ExpressionOperand: TExpressionOperand): Boolean;
    function ParseConstant(Block: TBlock; out ConstantRef: TConstant): Boolean;
    function ParseVariable(Block: TBlock; out VariableRef: TVariable): Boolean;
    function ParseIfThenElse(Block: TBlock; out IfThenElse: TIfThenElse): Boolean;
    function ParseWhileDo(Block: TBlock; out WhileDo: TWhileDo): Boolean;
    function ParseForToDo(Block: TBlock; out ForToDo: TForToDo): Boolean;
    procedure TokenizerError(Pos: TPoint; Text: string);
    procedure InitSystemBlock(Block: TBlock);
  public
    Source: string;
    Prog: TProgram;
    procedure Parse;
    procedure Error(Text: string);
    constructor Create;
    destructor Destroy; override;
    property OnError: TErrorEvent read FOnError write FOnError;
  end;


implementation

{ TParser }

function TParser.ParseBeginEnd(Block: TBlock; out BeginEnd: TBeginEnd): Boolean;
var
  Command: TCommand;
begin
  if Tokenizer.CheckNext('begin', tkKeyword) then begin
    BeginEnd := TBeginEnd.Create;
    Result := True;
    Tokenizer.Expect('begin', tkKeyword);
    while not Tokenizer.CheckNext('end', tkKeyword) do begin
      if ParseCommand(Block, Command) then begin
        BeginEnd.Commands.Add(Command);
        Tokenizer.Expect(';', tkSpecialSymbol);
      end else begin
        Error('Unexpected token ' + Tokenizer.GetNext.Text);
        Result := False;
        Break;
      end;
    end;
    Tokenizer.Expect('end', tkKeyword);
  end else Result := False;
end;

function TParser.ParseFunctionCall(Block: TBlock; out FunctionCall: TFunctionCall
  ): Boolean;
var
  Token: TToken;
  LastPos: TTokenizerPos;
  FunctionDef: TFunction;
  Expression: TExpression;
  I: Integer;
begin
  LastPos := Tokenizer.Pos;
  Token := Tokenizer.GetNext;
  if Token.Kind = tkIdentifier then begin
    FunctionDef := Block.GetFunction(Token.Text);
    if Assigned(FunctionDef) then begin
      FunctionCall := TFunctionCall.Create;
      FunctionCall.FunctionDef := FunctionDef;
      if Tokenizer.CheckNext('(', tkSpecialSymbol) then begin
        Tokenizer.Expect('(', tkSpecialSymbol);
        for I := 0 to FunctionDef.Params.Count - 1 do begin
          if I > 0 then Tokenizer.Expect(',', tkSpecialSymbol);
          if ParseExpression(Block, Expression) then begin
            if Expression.GetType = TFunctionParameter(FunctionDef.Params[I]).TypeRef then
              FunctionCall.Params.Add(Expression)
              else Error('Function parameter mismatch.');
          end else Error('Expected function parameter.');
        end;
        Tokenizer.Expect(')', tkSpecialSymbol);
      end;
      Result := True;
    end else begin
      Result := False;
      Tokenizer.Pos := LastPos;
    end;
  end else begin
    Result := False;
    Tokenizer.Pos := LastPos;
  end;
end;

function TParser.ParseCommand(Block: TBlock; out Command: TCommand): Boolean;
var
  BeginEnd: TBeginEnd;
  FunctionCall: TFunctionCall;
  Assignment: TAssignment;
  IfThenElse: TIfThenElse;
  WhileDo: TWhileDo;
  ForToDo: TForToDo;
begin
  if ParseIfThenElse(Block, IfThenElse) then begin
    Result := True;
    Command := IfThenElse;
  end else
  if ParseWhileDo(Block, WhileDo) then begin
    Result := True;
    Command := WhileDo;
  end else
  if ParseForToDo(Block, ForToDo) then begin
    Result := True;
    Command := ForToDo;
  end else
  if ParseBeginEnd(Block, BeginEnd) then begin
    Result := True;
    Command := BeginEnd;
  end else
  if ParseFunctionCall(Block, FunctionCall) then begin
    Result := True;
    Command := FunctionCall;
  end else
  if ParseAssignment(Block, Assignment) then begin
    Result := True;
    Command := Assignment;
  end else Result := False;
end;

function TParser.ParseProgram(SystemBlock: TBlock; out Prog: TProgram): Boolean;
var
  Block: TBlock;
  Token: TToken;
begin
  Result := False;
  Prog := TProgram.Create;
  Prog.SystemBlock.Free;
  Prog.SystemBlock := SystemBlock;
  if Tokenizer.CheckNext('program', tkKeyword) then begin
    Tokenizer.Expect('program', tkKeyword);
    Token := Tokenizer.GetNext;
    if Token.Kind = tkIdentifier then
      Prog.Name := Token.Text;
    Tokenizer.Expect(';', tkSpecialSymbol);
  end;
  if ParseBlock(SystemBlock, Block) then begin
    Result := True;
    Prog.Block.Free;
    Prog.Block := Block;
    Tokenizer.Expect('.', tkSpecialSymbol);
  end else begin
    FreeAndNil(Prog);
    Error('Expected begin-end block.');
  end;
end;

function TParser.ParseBlock(ParentBlock: TBlock; out Block: TBlock): Boolean;
var
  BeginEnd: TBeginEnd;
begin
  Result := False;
  Block := TBlock.Create;
  Block.Parent := ParentBlock;
  ParseBlockVar(Block);
  ParseBlockConst(Block);
  if ParseBeginEnd(Block, BeginEnd) then begin
    Result := True;
    Block.BeginEnd.Free;
    Block.BeginEnd := BeginEnd;
  end else Block.Free;
end;

function TParser.ParseBlockVar(Block: TBlock): Boolean;
var
  Token: TToken;
  Variable: TVariable;
  TypeRef: TType;
begin
  if Tokenizer.CheckNext('var', tkKeyword) then begin
    Result := True;
    Tokenizer.Expect('var', tkKeyword);
    while Tokenizer.CheckNextKind(tkIdentifier) do begin
      Token := Tokenizer.GetNext;
      if Token.Kind = tkIdentifier then begin
        Variable := Block.Variables.SearchByName(Token.Text);
        if not Assigned(Variable) then begin
          Variable := TVariable.Create;
          Variable.Name := Token.Text;
          Block.Variables.Add(Variable);
          Tokenizer.Expect(':', tkSpecialSymbol);
          Token := Tokenizer.GetNext;
          if Token.Kind = tkIdentifier then begin
            TypeRef := Block.GetType(Token.Text);
            if Assigned(TypeRef) then begin
              Variable.TypeRef := TypeRef;
            end else Error('Type ' + Token.Text + ' not found.');
          end;
        end else Error('Variable ' + Token.Text + ' redefined.');
        Tokenizer.Expect(';', tkSpecialSymbol);
      end else begin
        Error('Expected variable name but ' + Token.Text + ' found.');
        Break;
      end;
    end;
  end else Result := False;
end;

function TParser.ParseBlockConst(Block: TBlock): Boolean;
var
  Token: TToken;
  Constant: TConstant;
  TypeRef: TType;
begin
  if Tokenizer.CheckNext('const', tkKeyword) then begin
    Result := True;
    Tokenizer.Expect('const', tkKeyword);
    while Tokenizer.CheckNextKind(tkIdentifier) do begin
      Token := Tokenizer.GetNext;
      if Token.Kind = tkIdentifier then begin
        Constant := Block.Constants.SearchByName(Token.Text);
        if not Assigned(Constant) then begin
          Constant := TConstant.Create;
          Constant.Name := Token.Text;
          Block.Constants.Add(Constant);
          Tokenizer.Expect(':', tkSpecialSymbol);
          Token := Tokenizer.GetNext;
          if Token.Kind = tkIdentifier then begin
            TypeRef := Block.GetType(Token.Text);
            if Assigned(TypeRef) then begin
              Constant.TypeRef := TypeRef;
            end else Error('Type ' + Token.Text + ' not found.');
          end;
          Tokenizer.Expect('=', tkSpecialSymbol);
          Token := Tokenizer.GetNext;
          if Token.Kind = tkNumber then begin
            Constant.Value := TValueInteger.Create;
            TValueInteger(Constant.Value).Value := StrToInt(Token.Text);
          end else
          if Token.Kind = tkString then begin
            Constant.Value := TValueString.Create;
            TValueString(Constant.Value).Value := Token.Text;
          end else Error('Expected string or number.');
        end else Error('Constant ' + Token.Text + ' redefined.');
        Tokenizer.Expect(';', tkSpecialSymbol);
      end else begin
        Error('Expected constant name but ' + Token.Text + ' found.');
        Break;
      end;
    end;
  end else Result := False;
end;

function TParser.ParseAssignment(Block: TBlock; out Assignment: TAssignment): Boolean;
var
  Token: TToken;
  Variable: TVariable;
  Expression: TExpression;
begin
  Result := False;
  Token := Tokenizer.GetNext;
  if Token.Kind = tkIdentifier then begin
    Result := True;
    Variable := Block.GetVariable(Token.Text);
    if Assigned(Variable) then begin
      Result := True;
      Assignment := TAssignment.Create;
      Assignment.Variable := Variable;
      Tokenizer.Expect(':=', tkSpecialSymbol);
      if ParseExpression(Block, Expression) then begin
        if Expression.GetType = Variable.TypeRef then begin
          Assignment.Expression.Free;
          Assignment.Expression := Expression;
        end else begin
          Result := False;
          Error('Assignment type mismatch.');
        end;
      end;
      if not Result then Assignment.Free;
    end else Error('Variable ' + Token.Text + ' not defined.');
  end;
end;

function TParser.ParseExpression(Block: TBlock; out Expression: TExpression
  ): Boolean;
var
  ExpressionOperation: TExpressionOperation;
  ExpressionOperand: TExpressionOperand;
begin
  Result := False;
  if ParseExpressionOperation(Block, ExpressionOperation) then begin
    Result := True;
    Expression := ExpressionOperation;
  end else
  if ParseExpressionOperand(Block, ExpressionOperand) then begin
    Result := True;
    Expression := ExpressionOperand;
  end;
end;

function TParser.ParseExpressionOperation(Block: TBlock; out
  ExpressionOperation: TExpressionOperation): Boolean;
var
  Operand: TExpressionOperand;
  Token: TToken;
  Expression: TExpression;
  LastPos: TTokenizerPos;
begin
  Result := False;
  LastPos := Tokenizer.Pos;
  if ParseExpressionOperand(Block, Operand) then begin
    Token := Tokenizer.GetNext;
    if (Token.Kind = tkSpecialSymbol) and Tokenizer.IsOperator(Token.Text) then begin
      Result := True;
      ExpressionOperation := TExpressionOperation.Create;
      ExpressionOperation.TypeRef := Operand.GetType;
      if Token.Text = '+' then ExpressionOperation.Operation := eoAdd
      else if Token.Text = '-' then ExpressionOperation.Operation := eoSub
      else if Token.Text = '=' then ExpressionOperation.Operation := eoEqual
      else if Token.Text = '<>' then ExpressionOperation.Operation := eoNotEqual
      else Error('Unsupported operator ' + Token.Text);
      ExpressionOperation.Items.Add(Operand);
      if ParseExpression(Block, Expression) then begin
        if Expression.GetType = Operand.GetType then
          ExpressionOperation.Items.Add(Expression)
          else Error('Expression operands needs to be same type.');
      end else Error('Missing operand.');
    end;
  end;
  if not Result then Tokenizer.Pos := LastPos;
end;

function TParser.ParseExpressionOperand(Block: TBlock; out
  ExpressionOperand: TExpressionOperand): Boolean;
var
  Variable: TVariable;
  Constant: TConstant;
  FunctionCall: TFunctionCall;
begin
  Result := False;
  if ParseFunctionCall(Block, FunctionCall) then begin
    Result := True;
    ExpressionOperand := TExpressionOperand.Create;
    ExpressionOperand.FunctionCall := FunctionCall;
    ExpressionOperand.OperandType := otFunctionCall;
  end else
  if ParseConstant(Block, Constant) then begin
    Result := True;
    ExpressionOperand := TExpressionOperand.Create;
    ExpressionOperand.ConstantRef := Constant;
    ExpressionOperand.OperandType := otConstant;
  end else
  if ParseVariable(Block, Variable) then begin
    Result := True;
    ExpressionOperand := TExpressionOperand.Create;
    ExpressionOperand.VariableRef := Variable;
    ExpressionOperand.OperandType := otVariable;
  end else Error('Expected expression operand.');
end;

function TParser.ParseConstant(Block: TBlock; out ConstantRef: TConstant
  ): Boolean;
var
  LastPos: TTokenizerPos;
  Token: TToken;
begin
  Result := False;
  LastPos := Tokenizer.Pos;
  Token := Tokenizer.GetNext;
  if Token.Kind = tkIdentifier then begin;
    ConstantRef := Block.GetConstant(Token.Text);
    if Assigned(ConstantRef) then begin
      Result := True;
    end;
  end else
  if Token.Kind = tkNumber then begin
    Result := True;
    ConstantRef := Block.Constants.AddNew('_C' + IntToStr(Block.Constants.Count));
    ConstantRef.TypeRef := Block.GetType('Integer');
    ConstantRef.Value := TValueInteger.Create;
    TValueInteger(ConstantRef.Value).Value := StrToInt(Token.Text);
  end else
  if Token.Kind = tkString then begin
    Result := True;
    ConstantRef := Block.Constants.AddNew('_C' + IntToStr(Block.Constants.Count));
    ConstantRef.TypeRef := Block.GetType('string');
    ConstantRef.Value := TValueString.Create;
    TValueString(ConstantRef.Value).Value := Token.Text;
  end;
  if not Result then Tokenizer.Pos := LastPos;
end;

function TParser.ParseVariable(Block: TBlock; out VariableRef: TVariable
  ): Boolean;
var
  LastPos: TTokenizerPos;
  Token: TToken;
begin
  Result := False;
  LastPos := Tokenizer.Pos;
  Token := Tokenizer.GetNext;
  if Token.Kind = tkIdentifier then begin;
    VariableRef := Block.GetVariable(Token.Text);
    if Assigned(VariableRef) then begin
      Result := True;
    end;
  end;
  if not Result then Tokenizer.Pos := LastPos;
end;

function TParser.ParseIfThenElse(Block: TBlock; out IfThenElse: TIfThenElse
  ): Boolean;
var
  Expression: TExpression;
  Command: TCommand;
begin
  Result := False;
  if Tokenizer.CheckNext('if', tkKeyword) then begin
    Tokenizer.Expect('if', tkKeyword);
    Result := True;
    IfThenElse := TIfThenElse.Create;
    if ParseExpression(Block, Expression) then begin
      IfThenElse.Expression.Free;
      IfThenElse.Expression := Expression;
      Tokenizer.Expect('then', tkKeyword);
      if ParseCommand(Block, Command) then begin
        IfThenElse.CommandThen.Free;
        IfThenElse.CommandThen := Command;
        if Tokenizer.CheckNext('else', tkKeyword) then begin
          Tokenizer.Expect('else', tkKeyword);
          if ParseCommand(Block, Command) then begin
            IfThenElse.CommandElse.Free;
            IfThenElse.CommandElse := Command;
          end else Error('Expected command');
        end;
      end else Error('Expected command');
    end else Error('Expected expression');
  end;
end;

function TParser.ParseWhileDo(Block: TBlock; out WhileDo: TWhileDo): Boolean;
var
  Expression: TExpression;
  Command: TCommand;
begin
  Result := False;
  if Tokenizer.CheckNext('while', tkKeyword) then begin
    Tokenizer.Expect('while', tkKeyword);
    Result := True;
    WhileDo := TWhileDo.Create;
    if ParseExpression(Block, Expression) then begin
      WhileDo.Expression.Free;
      WhileDo.Expression := Expression;
      Tokenizer.Expect('do', tkKeyword);
      if ParseCommand(Block, Command) then begin
        WhileDo.Command.Free;
        WhileDo.Command := Command;
      end else Error('Expected command');
    end else Error('Expected expression');
  end;
end;

function TParser.ParseForToDo(Block: TBlock; out ForToDo: TForToDo): Boolean;
var
  Expression: TExpression;
  VariableRef: TVariable;
  Command: TCommand;
begin
  Result := False;
  if Tokenizer.CheckNext('for', tkKeyword) then begin
    Tokenizer.Expect('for', tkKeyword);
    Result := True;
    ForToDo := TForToDo.Create;
    if ParseVariable(Block, VariableRef) then begin
      ForToDo.VariableRef := VariableRef;
      Tokenizer.Expect(':=', tkSpecialSymbol);
      if ParseExpression(Block, Expression) then begin
        ForToDo.ExpressionFrom.Free;
        ForToDo.ExpressionFrom := Expression;
        Tokenizer.Expect('to', tkKeyword);
        if ParseExpression(Block, Expression) then begin
          ForToDo.ExpressionTo.Free;
          ForToDo.ExpressionTo := Expression;
          Tokenizer.Expect('do', tkKeyword);
          if ParseCommand(Block, Command) then begin
            ForToDo.Command.Free;
            ForToDo.Command := Command;
          end else Error('Expected command.');
        end else Error('Expected expression.');
      end else Error('Expected expression.');
    end else Error('Expected control variable.');
  end;
end;

procedure TParser.TokenizerError(Pos: TPoint; Text: string);
begin
  if Assigned(FOnError) then
    FOnError(Pos, Text);
end;

procedure TParser.InitSystemBlock(Block: TBlock);
var
  TypeBoolean: TType;
  TypeString: TType;
  TypeInteger: TType;
begin
  TypeBoolean := Block.Types.AddNew('Boolean');
  with TypeBoolean do begin
    ValueClass := TValueBoolean;
    with Functions.AddNew('_Assign') do begin
      Params.AddNew('Source', TypeBoolean);
      ResultType := TypeBoolean;
    end;
    with Functions.AddNew('_Equal') do begin
      Params.AddNew('A', TypeBoolean);
      Params.AddNew('B', TypeBoolean);
      ResultType := TypeBoolean;
    end;
  end;
  TypeString := Block.Types.AddNew('string');
  with TypeString do begin
    ValueClass := TValueString;
    with Functions.AddNew('_Assign') do begin
      Params.AddNew('Source', TypeString);
      ResultType := TypeString;
    end;
    with Functions.AddNew('_Add') do begin
      Params.AddNew('A', TypeString);
      Params.AddNew('B', TypeString);
      ResultType := TypeString;
    end;
    with Functions.AddNew('_Equal') do begin
      Params.AddNew('A', TypeString);
      Params.AddNew('B', TypeString);
      ResultType := TypeBoolean;
    end;
    with Functions.AddNew('_NotEqual') do begin
      Params.AddNew('A', TypeString);
      Params.AddNew('B', TypeString);
      ResultType := TypeBoolean;
    end;
  end;
  TypeInteger := Block.Types.AddNew('Integer');
  with TypeInteger do begin
    ValueClass := TValueInteger;
    with Functions.AddNew('_Assign') do begin
      Params.AddNew('Source', TypeInteger);
      ResultType := TypeInteger;
    end;
    with Functions.AddNew('_Add') do begin
      Params.AddNew('A', TypeInteger);
      Params.AddNew('B', TypeInteger);
      ResultType := TypeInteger;
    end;
    with Functions.AddNew('_Sub') do begin
      Params.AddNew('A', TypeInteger);
      Params.AddNew('B', TypeInteger);
      ResultType := TypeInteger;
    end;
    with Functions.AddNew('_Equal') do begin
      Params.AddNew('A', TypeInteger);
      Params.AddNew('B', TypeInteger);
      ResultType := TypeBoolean;
    end;
    with Functions.AddNew('_NotEqual') do begin
      Params.AddNew('A', TypeInteger);
      Params.AddNew('B', TypeInteger);
      ResultType := TypeBoolean;
    end;
  end;
  with Block.Functions.AddNew('IntToStr') do begin
    Params.AddNew('Value', TypeInteger);
    ResultType := TypeString;
  end;
  with Block.Functions.AddNew('StrToInt') do begin
    Params.AddNew('Value', TypeString);
    ResultType := TypeInteger;
  end;
  with Block.Functions.AddNew('WriteLn') do begin
    Params.AddNew('Text', TypeString);
  end;
  with Block.Functions.AddNew('Write') do begin
    Params.AddNew('Text', TypeString);
  end;
end;

procedure TParser.Parse;
var
  NewProg: TProgram;
  SystemBlock: TBlock;
begin
  Tokenizer.Source := Source;
  Tokenizer.Init;
  SystemBlock := TBlock.Create;
  InitSystemBlock(SystemBlock);
  if ParseProgram(SystemBlock, NewProg) then begin
    Prog := NewProg;
  end else Tokenizer.Error('Cannot parse program.');
end;

procedure TParser.Error(Text: string);
begin
  if Assigned(FOnError) then
    FOnError(Tokenizer.Pos.Pos, Text);
end;

constructor TParser.Create;
begin
  Tokenizer := TTokenizer.Create;
  Tokenizer.OnError := TokenizerError;
end;

destructor TParser.Destroy;
begin
  Tokenizer.Free;
  inherited Destroy;
end;

end.

