unit USource;

{$mode delphi}

interface

uses
  Classes, SysUtils, Contnrs;

type
  TExpressions = class;
  TFunctions = class;

  TValue = class
  end;

  TValueString = class(TValue)
    Value: string;
  end;

  TValueInteger = class(TValue)
    Value: Integer;
  end;

  TValueBoolean = class(TValue)
    Value: Boolean;
  end;

  TValueClass = class of TValue;

  { TType }

  TType = class
    Name: string;
    Functions: TFunctions;
    ValueClass: TValueClass;
    constructor Create;
    destructor Destroy; override;
  end;

  { TTypes }

  TTypes = class(TObjectList)
    function SearchByName(Name: string): TType;
    function AddNew(Name: string): TType;
  end;

  TVariable = class
    Name: string;
    TypeRef: TType;
  end;

  { TVariables }

  TVariables = class(TObjectList)
    function SearchByName(Name: string): TVariable;
  end;

  TConstant = class
    Name: string;
    TypeRef: TType;
    Value: TValue;
  end;

  { TConstants }

  TConstants = class(TObjectList)
    function SearchByName(Name: string): TConstant;
    function AddNew(Name: string): TConstant;
  end;

  TFunctionParameter = class
    Name: string;
    TypeRef: TType;
  end;

  { TFunctionParameters }

  TFunctionParameters = class(TObjectList)
    function SearchByName(Name: string): TFunctionParameter;
    function AddNew(Name: string; TypeRef: TType): TFunctionParameter;
  end;

  { TFunction }

  TFunction = class
    Name: string;
    Params: TFunctionParameters;
    ResultType: TType;
    constructor Create;
    destructor Destroy; override;
  end;

  { TFunctions }

  TFunctions = class(TObjectList)
    function SearchByName(Name: string): TFunction;
    function AddNew(Name: string): TFunction;
  end;

  TCommand = class
  end;

  { TFunctionCall }

  TFunctionCall = class(TCommand)
    FunctionDef: TFunction;
    Params: TExpressions;
    constructor Create;
    destructor Destroy; override;
  end;

  { TBeginEnd }

  TBeginEnd = class(TCommand)
    Commands: TObjectList;
    procedure Clear;
    constructor Create;
    destructor Destroy; override;
  end;

  TExpressionOperator = (eoAdd, eoSub, eoMultiply, eoDivide, eoModulo, eoAnd, eoXor,
    eoOr, eoShl, eoShr, eoEqual, eoNotEqual);

  { TExpression }

  TExpression = class
    function GetType: TType; virtual;
  end;

  { TExpressionOperation }

  TExpressionOperation = class(TExpression)
    TypeRef: TType;
    Operation: TExpressionOperator;
    Items: TExpressions;
    constructor Create;
    destructor Destroy; override;
    function GetType: TType; override;
  end;

  TExpressionOperandType = (otVariable, otConstant, otFunctionCall);

  { TExpressionOperand }

  TExpressionOperand = class(TExpression)
    OperandType: TExpressionOperandType;
    VariableRef: TVariable;
    ConstantRef: TConstant;
    FunctionCall: TFunctionCall;
    function GetType: TType; override;
  end;

  TExpressions = class(TObjectList)
  end;

  { TAssignment }

  TAssignment = class(TCommand)
    Variable: TVariable;
    Expression: TExpression;
    constructor Create;
    destructor Destroy; override;
  end;

  { TIfThenElse }

  TIfThenElse = class(TCommand)
    Expression: TExpression;
    CommandThen: TCommand;
    CommandElse: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  { TWhileDo }

  TWhileDo = class(TCommand)
    Expression: TExpression;
    Command: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  { TForToDo }

  TForToDo = class(TCommand)
    VariableRef: TVariable;
    ExpressionFrom: TExpression;
    ExpressionTo: TExpression;
    Command: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  { TBlock }

  TBlock = class
    Parent: TBlock;
    Variables: TVariables;
    Constants: TConstants;
    Functions: TFunctions;
    Types: TTypes;
    BeginEnd: TBeginEnd;
    procedure Clear;
    function GetType(Name: string): TType;
    function GetConstant(Name: string): TConstant;
    function GetVariable(Name: string): TVariable;
    function GetFunction(Name: string): TFunction;
    constructor Create;
    destructor Destroy; override;
  end;

  { TProgram }

  TProgram = class
    Name: string;
    SystemBlock: TBlock;
    Block: TBlock;
    procedure Clear;
    constructor Create;
    destructor Destroy; override;
  end;

implementation

{ TForToDo }

constructor TForToDo.Create;
begin
  ExpressionFrom := TExpression.Create;
  ExpressionTo := TExpression.Create;
  Command := TCommand.Create;
end;

destructor TForToDo.Destroy;
begin
  Command.Free;
  ExpressionTo.Free;
  ExpressionFrom.Free;
  inherited Destroy;
end;

{ TExpression }

function TExpression.GetType: TType;
begin
  Result := nil;
end;

{ TExpressionOperand }

function TExpressionOperand.GetType: TType;
begin
  if OperandType = otFunctionCall then Result := FunctionCall.FunctionDef.ResultType
  else if OperandType = otConstant then Result := ConstantRef.TypeRef
  else if OperandType = otVariable then Result := VariableRef.TypeRef
  else raise Exception.Create('Unsupported operand type');
end;

{ TFunctionParameters }

function TFunctionParameters.SearchByName(Name: string): TFunctionParameter;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TFunctionParameter(Items[I]).Name <> Name) do Inc(I);
  if I < Count then Result := TFunctionParameter(Items[I])
    else Result := nil;
end;

function TFunctionParameters.AddNew(Name: string; TypeRef: TType): TFunctionParameter;
begin
  Result := TFunctionParameter.Create;
  Result.Name := Name;
  Result.TypeRef := TypeRef;
  Add(Result);
end;

{ TFunction }

constructor TFunction.Create;
begin
  Params := TFunctionParameters.Create;
end;

destructor TFunction.Destroy;
begin
  Params.Free;
  inherited Destroy;
end;

{ TType }

constructor TType.Create;
begin
  Functions := TFunctions.Create;
end;

destructor TType.Destroy;
begin
  Functions.Free;
  inherited Destroy;
end;

{ TTypes }

function TTypes.SearchByName(Name: string): TType;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TType(Items[I]).Name <> Name) do Inc(I);
  if I < Count then Result := TType(Items[I])
    else Result := nil;
end;

function TTypes.AddNew(Name: string): TType;
begin
  Result := TType.Create;
  Result.Name := Name;
  Add(Result);
end;

{ TExpressionOperation }

constructor TExpressionOperation.Create;
begin
  Items := TExpressions.Create;
end;

destructor TExpressionOperation.Destroy;
begin
  Items.Free;
  inherited Destroy;
end;

function TExpressionOperation.GetType: TType;
begin
  Result := TypeRef;
end;

{ TAssignment }

constructor TAssignment.Create;
begin
  Variable := nil;
  Expression := TExpression.Create;
end;

destructor TAssignment.Destroy;
begin
  Variable := nil;
  Expression.Free;
  inherited Destroy;
end;

{ TIfThenElse }

constructor TIfThenElse.Create;
begin
  Expression := TExpression.Create;
  CommandThen := TCommand.Create;
  CommandElse := TCommand.Create;
end;

destructor TIfThenElse.Destroy;
begin
  Expression.Free;
  CommandThen.Free;
  CommandElse.Free;
  inherited Destroy;
end;

{ TWhileDo }

constructor TWhileDo.Create;
begin
  Expression := TExpression.Create;
  Command := TCommand.Create;
end;

destructor TWhileDo.Destroy;
begin
  Expression.Free;
  Command.Free;
  inherited Destroy;
end;

{ TFunctionCall }

constructor TFunctionCall.Create;
begin
  Params := TExpressions.Create;
end;

destructor TFunctionCall.Destroy;
begin
  Params.Free;
  inherited Destroy;
end;

{ TFunctions }

function TFunctions.SearchByName(Name: string): TFunction;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TFunction(Items[I]).Name <> Name) do Inc(I);
  if I < Count then Result := TFunction(Items[I])
    else Result := nil;
end;

function TFunctions.AddNew(Name: string): TFunction;
begin
  Result := TFunction.Create;
  Result.Name := Name;
  Add(Result);
end;

{ TConstants }

function TConstants.SearchByName(Name: string): TConstant;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TConstant(Items[I]).Name <> Name) do Inc(I);
  if I < Count then Result := TConstant(Items[I])
    else Result := nil;
end;

function TConstants.AddNew(Name: string): TConstant;
begin
  Result := TConstant.Create;
  Result.Name := Name;
  Add(Result);
end;

{ TVariables }

function TVariables.SearchByName(Name: string): TVariable;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TVariable(Items[I]).Name <> Name) do Inc(I);
  if I < Count then Result := TVariable(Items[I])
    else Result := nil;
end;

{ TBlock }

procedure TBlock.Clear;
begin
  Functions.Clear;
  Constants.Clear;
  Variables.Clear;
  Types.Clear;
end;

function TBlock.GetType(Name: string): TType;
begin
  Result := Types.SearchByName(Name);
  if not Assigned(Result) and Assigned(Parent) then
    Result := Parent.Types.SearchByName(Name);
end;

function TBlock.GetConstant(Name: string): TConstant;
begin
  Result := Constants.SearchByName(Name);
  if not Assigned(Result) and Assigned(Parent) then
    Result := Parent.Constants.SearchByName(Name);
end;

function TBlock.GetVariable(Name: string): TVariable;
begin
  Result := Variables.SearchByName(Name);
  if not Assigned(Result) and Assigned(Parent) then
    Result := Parent.Variables.SearchByName(Name);
end;

function TBlock.GetFunction(Name: string): TFunction;
begin
  Result := Functions.SearchByName(Name);
  if not Assigned(Result) and Assigned(Parent) then
    Result := Parent.Functions.SearchByName(Name);
end;

constructor TBlock.Create;
begin
  Constants := TConstants.Create;
  Variables := TVariables.Create;
  Functions := TFunctions.Create;
  Types := TTypes.Create;
  BeginEnd := TBeginEnd.Create;
end;

destructor TBlock.Destroy;
begin
  BeginEnd.Free;
  Types.Free;
  Variables.Free;
  Constants.Free;
  Functions.Free;
  inherited Destroy;
end;

{ TBeginEnd }

procedure TBeginEnd.Clear;
begin
  Commands.Clear;
end;

constructor TBeginEnd.Create;
begin
  Commands := TObjectList.Create;
end;

destructor TBeginEnd.Destroy;
begin
  Commands.Free;
  inherited Destroy;
end;

{ TProgram }

procedure TProgram.Clear;
begin
  Name := '';
end;

constructor TProgram.Create;
begin
  SystemBlock := TBlock.Create;
  Block := TBlock.Create;
end;

destructor TProgram.Destroy;
begin
  Block.Free;
  SystemBlock.Free;
  inherited Destroy;
end;



end.

